@@ -2,7 +2,6 @@
 # Copyright (c) 2022 Aakash Sen Sharma
 
 import importlib.util
-import os
 import sys
 from pathlib import Path
 
@@ -16,10 +15,8 @@ assert include_dir.is_dir(), f"missing {include_dir}"
 
 def load_version():
     """Load the current pywlroots version"""
-    dirname = os.path.dirname(__file__)
-    spec = importlib.util.spec_from_file_location(
-        "wlroots.version", os.path.join(dirname, "version.py")
-    )
+    file_location = Path(__file__).parent / "version.py"
+    spec = importlib.util.spec_from_file_location("wlroots.version", str(file_location))
     assert spec is not None
     assert spec.loader is not None
     version_module = importlib.util.module_from_spec(spec)
@@ -37,7 +34,7 @@ def load_wlroots_version():
 
     try:
         lib = ffi.verify("#include <wlr/version.h>")
-    except (PermissionError, OSError):
+    except (PermissionError, OSError, VerificationError):
         lib = importlib.import_module("wlroots").lib
 
     return f"{lib.WLR_VERSION_MAJOR}.{lib.WLR_VERSION_MINOR}.{lib.WLR_VERSION_MICRO}"  # type: ignore[attr-defined]
@@ -64,10 +61,17 @@ def has_xwayland() -> bool:
     Check for XWayland headers. If present, wlroots was built with XWayland support, so
     pywlroots can be too.
     """
+    file_location = Path(__file__).parent / "_build.py"
+    spec = importlib.util.spec_from_file_location("wlroots._build", str(file_location))
+    assert spec is not None
+    assert spec.loader is not None
+    build_module = importlib.util.module_from_spec(spec)
     try:
-        return importlib.import_module("wlroots._build").has_xwayland
-    except ModuleNotFoundError:
+        spec.loader.exec_module(build_module)
+    except FileNotFoundError:
         pass
+    else:
+        return build_module.has_xwayland
 
     try:
         FFI().verify(
@@ -83,7 +87,7 @@ def has_xwayland() -> bool:
         )
         has_xwayland = False
 
-    with (Path(__file__).parent / "_build.py").open("w") as f:
+    with file_location.open("w") as f:
         f.write(f"has_xwayland = {has_xwayland}\n")
 
     return has_xwayland
@@ -117,12 +121,17 @@ void wlr_backend_destroy(struct wlr_backend *backend);
 struct wlr_session *wlr_backend_get_session(struct wlr_backend *backend);
 """
 
-# backend/libinput.h
+# wlr/backend/libinput.h
 CDEF += """
 struct libinput_device *wlr_libinput_get_device_handle(struct wlr_input_device *dev);
 bool wlr_input_device_is_libinput(struct wlr_input_device *device);
 """
 
+# wlr/backend/multi.h
+CDEF += """
+bool wlr_backend_is_multi(struct wlr_backend *backend);
+"""
+
 # backend/session.h
 CDEF += """
 bool wlr_session_change_vt(struct wlr_session *session, unsigned vt);
@@ -294,6 +303,8 @@ void wlr_cursor_set_surface(struct wlr_cursor *cur, struct wlr_surface *surface,
     int32_t hotspot_x, int32_t hotspot_y);
 void wlr_cursor_attach_input_device(struct wlr_cursor *cur,
     struct wlr_input_device *dev);
+void wlr_cursor_detach_input_device(struct wlr_cursor *cur,
+    struct wlr_input_device *dev);
 void wlr_cursor_attach_output_layout(struct wlr_cursor *cur,
     struct wlr_output_layout *l);
 void wlr_cursor_map_to_output(struct wlr_cursor *cur,
@@ -1076,6 +1087,17 @@ struct wlr_output_mode {
 };
 
 struct wlr_output_state {
+    bool enabled;
+    float scale;
+    enum wl_output_transform transform;
+    bool adaptive_sync_enabled;
+    uint32_t render_format;
+    enum wl_output_subpixel subpixel;
+    struct wlr_output_mode *mode;
+    struct {
+        int32_t width, height;
+        int32_t refresh; // mHz, may be zero
+    } custom_mode;
     ...;
 };
 
@@ -1153,6 +1175,7 @@ void wlr_output_set_custom_mode(struct wlr_output *output, int32_t width,
     int32_t height, int32_t refresh);
 void wlr_output_set_transform(struct wlr_output *output,
     enum wl_output_transform transform);
+void wlr_output_enable_adaptive_sync(struct wlr_output *output, bool enabled);
 void wlr_output_set_scale(struct wlr_output *output, float scale);
 
 bool wlr_output_attach_render(struct wlr_output *output, int *buffer_age);
@@ -1165,12 +1188,36 @@ void wlr_output_set_damage(struct wlr_output *output,
 bool wlr_output_test(struct wlr_output *output);
 bool wlr_output_commit(struct wlr_output *output);
 void wlr_output_rollback(struct wlr_output *output);
+bool wlr_output_test_state(struct wlr_output *output,
+    const struct wlr_output_state *state);
+bool wlr_output_commit_state(struct wlr_output *output,
+    const struct wlr_output_state *state);
+
+void wlr_output_state_set_enabled(struct wlr_output_state *state,
+    bool enabled);
+void wlr_output_state_set_mode(struct wlr_output_state *state,
+    struct wlr_output_mode *mode);
+void wlr_output_state_set_custom_mode(struct wlr_output_state *state,
+    int32_t width, int32_t height, int32_t refresh);
+void wlr_output_state_set_scale(struct wlr_output_state *state, float scale);
+void wlr_output_state_set_transform(struct wlr_output_state *state,
+    enum wl_output_transform transform);
+void wlr_output_state_set_adaptive_sync_enabled(struct wlr_output_state *state,
+    bool enabled);
+void wlr_output_state_set_render_format(struct wlr_output_state *state,
+    uint32_t format);
+void wlr_output_state_set_subpixel(struct wlr_output_state *state,
+    enum wl_output_subpixel subpixel);
 
 void wlr_output_render_software_cursors(struct wlr_output *output,
     struct pixman_region32 *damage);
 
 enum wl_output_transform wlr_output_transform_invert(
     enum wl_output_transform tr);
+
+enum wl_output_transform wlr_output_transform_compose(
+    enum wl_output_transform tr_a, enum wl_output_transform tr_b);
+
 """
 
 # types/wlr_output_damage.h
@@ -1334,6 +1381,10 @@ struct wlr_output_configuration_head_v1 *
     struct wlr_output_configuration_v1 *config, struct wlr_output *output);
 void wlr_output_configuration_v1_destroy(
     struct wlr_output_configuration_v1 *config);
+
+void wlr_output_head_v1_state_apply(
+    const struct wlr_output_head_v1_state *head_state,
+    struct wlr_output_state *output_state);
 """
 
 # types/wlr_output_powewr_management_v1.h
@@ -2025,7 +2076,18 @@ struct wlr_seat_keyboard_state {
     } events;
     ...;
 };
-struct wlr_seat_touch_state { ...; };
+
+struct wlr_seat_touch_state {
+    struct wlr_seat *seat;
+    struct wl_list touch_points; // wlr_touch_point.link
+
+    uint32_t grab_serial;
+    uint32_t grab_id;
+
+    struct wlr_seat_touch_grab *grab;
+    struct wlr_seat_touch_grab *default_grab;
+    ...;
+};
 
 struct wlr_seat {
     struct wl_global *global;
@@ -2158,13 +2220,37 @@ bool wlr_seat_pointer_has_grab(struct wlr_seat *seat);
 void wlr_seat_set_keyboard(struct wlr_seat *seat, struct wlr_keyboard *keyboard);
 struct wlr_keyboard *wlr_seat_get_keyboard(struct wlr_seat *seat);
 
+struct wlr_touch_point *wlr_seat_touch_get_point(struct wlr_seat *seat,
+        int32_t touch_id);
+void wlr_seat_touch_point_focus(struct wlr_seat *seat,
+        struct wlr_surface *surface, uint32_t time_msec,
+        int32_t touch_id, double sx, double sy);
+void wlr_seat_touch_point_clear_focus(struct wlr_seat *seat, uint32_t time_msec,
+        int32_t touch_id);
+uint32_t wlr_seat_touch_send_down(struct wlr_seat *seat,
+        struct wlr_surface *surface, uint32_t time_msec,
+        int32_t touch_id, double sx, double sy);
+void wlr_seat_touch_send_up(struct wlr_seat *seat, uint32_t time_msec,
+        int32_t touch_id);
+void wlr_seat_touch_send_motion(struct wlr_seat *seat, uint32_t time_msec,
+        int32_t touch_id, double sx, double sy);
+void wlr_seat_touch_send_cancel(struct wlr_seat *seat, struct wlr_surface *surface);
+void wlr_seat_touch_send_frame(struct wlr_seat *seat);
 uint32_t wlr_seat_touch_notify_down(struct wlr_seat *seat,
-    struct wlr_surface *surface, uint32_t time_msec,
-    int32_t touch_id, double sx, double sy);
+        struct wlr_surface *surface, uint32_t time_msec,
+        int32_t touch_id, double sx, double sy);
 void wlr_seat_touch_notify_up(struct wlr_seat *seat, uint32_t time_msec,
-    int32_t touch_id);
+        int32_t touch_id);
 void wlr_seat_touch_notify_motion(struct wlr_seat *seat, uint32_t time_msec,
-    int32_t touch_id, double sx, double sy);
+        int32_t touch_id, double sx, double sy);
+void wlr_seat_touch_notify_cancel(struct wlr_seat *seat,
+        struct wlr_surface *surface);
+void wlr_seat_touch_notify_frame(struct wlr_seat *seat);
+int wlr_seat_touch_num_points(struct wlr_seat *seat);
+void wlr_seat_touch_start_grab(struct wlr_seat *wlr_seat,
+        struct wlr_seat_touch_grab *grab);
+void wlr_seat_touch_end_grab(struct wlr_seat *wlr_seat);
+bool wlr_seat_touch_has_grab(struct wlr_seat *seat);
 
 void wlr_seat_keyboard_start_grab(struct wlr_seat *wlr_seat,
     struct wlr_seat_keyboard_grab *grab);
@@ -2181,6 +2267,8 @@ void wlr_seat_pointer_notify_clear_focus(struct wlr_seat *wlr_seat);
 bool wlr_seat_keyboard_has_grab(struct wlr_seat *seat);
 bool wlr_seat_validate_pointer_grab_serial(struct wlr_seat *seat,
     struct wlr_surface *origin, uint32_t serial);
+
+bool wlr_surface_accepts_touch(struct wlr_seat *wlr_seat, struct wlr_surface *surface);
 """
 
 # types/wlr_server_decoration.h
@@ -2209,6 +2297,167 @@ void wlr_server_decoration_manager_set_default_mode(
     struct wlr_server_decoration_manager *manager, uint32_t default_mode);
 """
 
+# types/wlr_single_pixel_buffer_v1.h
+CDEF += """
+struct wlr_single_pixel_buffer_manager_v1 *wlr_single_pixel_buffer_manager_v1_create(
+    struct wl_display *display);
+"""
+
+# types/wlr_session_lock_v1.h
+CDEF += """
+struct wlr_session_lock_manager_v1 {
+    struct wl_global *global;
+
+    struct {
+        struct wl_signal new_lock; // struct wlr_session_lock_v1 *
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_session_lock_v1 {
+    struct wl_resource *resource;
+
+    struct wl_list surfaces; // struct wlr_session_lock_surface_v1.link
+
+    struct {
+        struct wl_signal new_surface; // struct wlr_session_lock_surface_v1 *
+        struct wl_signal unlock;
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+};
+
+struct wlr_session_lock_surface_v1 {
+    struct wl_resource *resource;
+    struct wl_list link; // wlr_session_lock_v1.surfaces
+
+    struct wlr_output *output;
+    struct wlr_surface *surface;
+
+    bool configured, mapped;
+
+    struct {
+        struct wl_signal map;
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+
+    ...;
+};
+
+struct wlr_session_lock_manager_v1 *wlr_session_lock_manager_v1_create(
+    struct wl_display *display);
+
+void wlr_session_lock_v1_send_locked(struct wlr_session_lock_v1 *lock);
+void wlr_session_lock_v1_destroy(struct wlr_session_lock_v1 *lock);
+
+uint32_t wlr_session_lock_surface_v1_configure(
+    struct wlr_session_lock_surface_v1 *lock_surface,
+    uint32_t width, uint32_t height);
+
+bool wlr_surface_is_session_lock_surface_v1(struct wlr_surface *surface);
+
+struct wlr_session_lock_surface_v1 *wlr_session_lock_surface_v1_from_wlr_surface(
+    struct wlr_surface *surface);
+"""
+
+
+# types/wlr_switch.h
+CDEF += """
+struct wlr_switch {
+    struct wlr_input_device base;
+
+    struct {
+        struct wl_signal toggle;
+    } events;
+
+    void *data;
+
+    ...;
+};
+
+enum wlr_switch_type {
+    WLR_SWITCH_TYPE_LID,
+    WLR_SWITCH_TYPE_TABLET_MODE,
+};
+
+enum wlr_switch_state {
+    WLR_SWITCH_STATE_OFF = 0,
+    WLR_SWITCH_STATE_ON,
+};
+
+struct wlr_switch_toggle_event {
+    uint32_t time_msec;
+    enum wlr_switch_type switch_type;
+    enum wlr_switch_state switch_state;
+};
+
+struct wlr_switch *wlr_switch_from_input_device(
+    struct wlr_input_device *input_device);
+"""
+
+
+# types/wlr_touch.h
+CDEF += """
+struct wlr_touch {
+    struct wlr_input_device base;
+    const struct wlr_touch_impl *impl;
+    char *output_name;
+    double width_mm, height_mm;
+    struct {
+        struct wl_signal down; // struct wlr_touch_down_event
+        struct wl_signal up; // struct wlr_touch_up_event
+        struct wl_signal motion; // struct wlr_touch_motion_event
+        struct wl_signal cancel; // struct wlr_touch_cancel_event
+        struct wl_signal frame;
+    } events;
+
+    void *data;
+
+    ...;
+};
+
+struct wlr_touch_down_event {
+    struct wlr_touch *touch;
+    uint32_t time_msec;
+    int32_t touch_id;
+    // From 0..1
+    double x, y;
+    ...;
+};
+
+struct wlr_touch_up_event {
+    struct wlr_touch *touch;
+    uint32_t time_msec;
+    int32_t touch_id;
+    ...;
+};
+
+struct wlr_touch_motion_event {
+    struct wlr_touch *touch;
+    uint32_t time_msec;
+    int32_t touch_id;
+    // From 0..1
+    double x, y;
+    ...;
+};
+
+struct wlr_touch_cancel_event {
+    struct wlr_touch *touch;
+    uint32_t time_msec;
+    int32_t touch_id;
+    ...;
+};
+
+struct wlr_touch *wlr_touch_from_input_device(
+    struct wlr_input_device *input_device);
+"""
+
 # types/wlr_virtual_keyboard_v1.h
 CDEF += """
 struct wlr_virtual_keyboard_manager_v1 {
@@ -2779,6 +3028,8 @@ SOURCE = """
 #include <wlr/backend.h>
 #include <wlr/backend/headless.h>
 #include <wlr/backend/libinput.h>
+#include <wlr/backend/multi.h>
+#include <wlr/interfaces/wlr_keyboard.h>
 #include <wlr/render/allocator.h>
 #include <wlr/render/drm_format_set.h>
 #include <wlr/render/wlr_renderer.h>
@@ -2817,6 +3068,10 @@ SOURCE = """
 #include <wlr/types/wlr_seat.h>
 #include <wlr/types/wlr_subcompositor.h>
 #include <wlr/types/wlr_server_decoration.h>
+#include <wlr/types/wlr_session_lock_v1.h>
+#include <wlr/types/wlr_single_pixel_buffer_v1.h>
+#include <wlr/types/wlr_switch.h>
+#include <wlr/types/wlr_touch.h>
 #include <wlr/types/wlr_virtual_keyboard_v1.h>
 #include <wlr/types/wlr_virtual_pointer_v1.h>
 #include <wlr/types/wlr_xcursor_manager.h>
